@@ -109,7 +109,6 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
         MVMint32 num_attrs        = MVM_repr_elems(tc, flat_list);
         MVMint32 info_alloc       = num_attrs == 0 ? 1 : num_attrs;
         MVMint32 cur_obj_attr     = 0;
-        MVMint32 cur_str_attr     = 0;
         MVMint32 cur_init_slot    = 0;
         MVMint32 i;
 
@@ -217,7 +216,9 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
                 }
                 else {
                     MVM_exception_throw_adhoc(tc,
-                        "CUnion representation only handles int, num, CArray, CPointer, CStruct, CPPStruct and CUnion");
+                        "CUnion representation only handles attributes of type:\n"
+                        "  (u)int8, (u)int16, (u)int32, (u)int64, (u)long, (u)longlong, num16, num32, (s)size_t, bool, Str\n"
+                        "  and types with representation: CArray, CPointer, CStruct, CPPStruct and CUnion");
                 }
             }
             else {
@@ -400,47 +401,51 @@ static void get_attribute(MVMThreadContext *tc, MVMSTable *st, MVMObject *root,
                 MVMObject *obj     = body->child_objs[real_slot];
                 if (!obj) {
                     /* No cached object. */
-                    void *cobj = get_ptr_at_offset(body->cunion, repr_data->struct_offsets[slot]);
-                    if (cobj) {
-                        MVMObject **child_objs = body->child_objs;
-                        if (type == MVM_CUNION_ATTR_CARRAY) {
-                            obj = MVM_nativecall_make_carray(tc, typeobj, cobj);
+                    if (repr_data->attribute_locations[slot] & MVM_CUNION_ATTR_INLINED) {
+                        if (type == MVM_CUNION_ATTR_CSTRUCT) {
+                            obj = MVM_nativecall_make_cstruct(tc, typeobj,
+                                (char *)body->cunion + repr_data->struct_offsets[slot]);
                         }
-                        else if(type == MVM_CUNION_ATTR_CSTRUCT) {
-                            if (repr_data->attribute_locations[slot] & MVM_CUNION_ATTR_INLINED)
-                                obj = MVM_nativecall_make_cstruct(tc, typeobj,
-                                    (char *)body->cunion + repr_data->struct_offsets[slot]);
-                            else
-                                obj = MVM_nativecall_make_cstruct(tc, typeobj, cobj);
+                        else if (type == MVM_CUNION_ATTR_CPPSTRUCT) {
+                            obj = MVM_nativecall_make_cppstruct(tc, typeobj,
+                                (char *)body->cunion + repr_data->struct_offsets[slot]);
                         }
-                        else if(type == MVM_CUNION_ATTR_CPPSTRUCT) {
-                            if (repr_data->attribute_locations[slot] & MVM_CUNION_ATTR_INLINED)
-                                obj = MVM_nativecall_make_cppstruct(tc, typeobj,
-                                    (char *)body->cunion + repr_data->struct_offsets[slot]);
-                            else
-                                obj = MVM_nativecall_make_cppstruct(tc, typeobj, cobj);
+                        else if (type == MVM_CUNION_ATTR_CUNION) {
+                            obj = MVM_nativecall_make_cunion(tc, typeobj,
+                                (char *)body->cunion + repr_data->struct_offsets[slot]);
                         }
-                        else if(type == MVM_CUNION_ATTR_CUNION) {
-                            if (repr_data->attribute_locations[slot] & MVM_CUNION_ATTR_INLINED)
-                                obj = MVM_nativecall_make_cunion(tc, typeobj,
-                                    (char *)body->cunion + repr_data->struct_offsets[slot]);
-                            else
-                                obj = MVM_nativecall_make_cunion(tc, typeobj, cobj);
-                        }
-                        else if(type == MVM_CUNION_ATTR_CPTR) {
-                            obj = MVM_nativecall_make_cpointer(tc, typeobj, cobj);
-                        }
-                        else if(type == MVM_CUNION_ATTR_STRING) {
-                            MVMROOT(tc, typeobj, {
-                                MVMString *str = MVM_string_utf8_decode(tc, tc->instance->VMString,
-                                    cobj, strlen(cobj));
-                                obj = MVM_repr_box_str(tc, typeobj, str);
-                            });
-                        }
-                        child_objs[real_slot] = obj;
                     }
                     else {
-                        obj = typeobj;
+                        void *cobj = get_ptr_at_offset(body->cunion, repr_data->struct_offsets[slot]);
+                        if (cobj) {
+                            MVMObject **child_objs = body->child_objs;
+                            if (type == MVM_CUNION_ATTR_CARRAY) {
+                                obj = MVM_nativecall_make_carray(tc, typeobj, cobj);
+                            }
+                            else if(type == MVM_CUNION_ATTR_CSTRUCT) {
+                                obj = MVM_nativecall_make_cstruct(tc, typeobj, cobj);
+                            }
+                            else if(type == MVM_CUNION_ATTR_CPPSTRUCT) {
+                                obj = MVM_nativecall_make_cppstruct(tc, typeobj, cobj);
+                            }
+                            else if(type == MVM_CUNION_ATTR_CUNION) {
+                                obj = MVM_nativecall_make_cunion(tc, typeobj, cobj);
+                            }
+                            else if(type == MVM_CUNION_ATTR_CPTR) {
+                                obj = MVM_nativecall_make_cpointer(tc, typeobj, cobj);
+                            }
+                            else if(type == MVM_CUNION_ATTR_STRING) {
+                                MVMROOT(tc, typeobj, {
+                                    MVMString *str = MVM_string_utf8_decode(tc, tc->instance->VMString,
+                                        cobj, strlen(cobj));
+                                    obj = MVM_repr_box_str(tc, typeobj, str);
+                                });
+                            }
+                            child_objs[real_slot] = obj;
+                        }
+                        else {
+                            obj = typeobj;
+                        }
                     }
                 }
                 result_reg->o = obj;
