@@ -288,6 +288,7 @@ static void deserialize_sc_deps(MVMThreadContext *tc, MVMCompUnit *cu, ReaderSta
         if (scb && scb->sc) {
             cu_body->scs_to_resolve[i] = NULL;
             MVM_ASSIGN_REF(tc, &(cu->common.header), cu_body->scs[i], scb->sc);
+            scb->claimed = 1;
         }
         else {
             if (!scb) {
@@ -666,6 +667,12 @@ void MVM_bytecode_finish_frame(MVMThreadContext *tc, MVMCompUnit *cu,
     for (j = 0; j < slvs; j++) {
         MVMuint16 lex_idx = read_int16(pos, 0);
         MVMuint16 flags   = read_int16(pos, 2);
+
+        if (lex_idx >= sf->body.num_lexicals) {
+            MVM_reentrantmutex_unlock(tc, (MVMReentrantMutex *)cu->body.update_mutex);
+            MVM_exception_throw_adhoc(tc, "Lexical index out of bounds: %d > %d", lex_idx, sf->body.num_lexicals);
+        }
+
         sf->body.static_env_flags[lex_idx] = flags;
         if (flags == 2 && !dump_only) {
             /* State variable; need to resolve wval immediately. Other kinds
