@@ -23,6 +23,7 @@ ffi_type * MVM_nativecall_get_ffi_type(MVMThreadContext *tc, MVMuint64 type_id)
         case MVM_NATIVECALL_ARG_ASCIISTR:
         case MVM_NATIVECALL_ARG_UTF8STR:
         case MVM_NATIVECALL_ARG_UTF16STR:
+        case MVM_NATIVECALL_ARG_CPPSTRUCT:
         case MVM_NATIVECALL_ARG_CSTRUCT:
         case MVM_NATIVECALL_ARG_CPOINTER:
         case MVM_NATIVECALL_ARG_CARRAY:
@@ -305,24 +306,29 @@ static void callback_handler(ffi_cif *cif, void *cb_result, void **cb_args, void
         MVMuint8 **backup_interp_bytecode_start = tc->interp_bytecode_start;
         MVMRegister **backup_interp_reg_base    = tc->interp_reg_base;
         MVMCompUnit **backup_interp_cu          = tc->interp_cu;
-        MVMFrame *backup_cur_frame              = tc->cur_frame;
+        MVMFrame *backup_cur_frame              = MVM_frame_force_to_heap(tc, tc->cur_frame);
         MVMFrame *backup_thread_entry_frame     = tc->thread_entry_frame;
-        MVMuint32 backup_mark                   = MVM_gc_root_temp_mark(tc);
-        jmp_buf backup_interp_jump;
-        memcpy(backup_interp_jump, tc->interp_jump, sizeof(jmp_buf));
+        MVMROOT(tc, backup_cur_frame, {
+        MVMROOT(tc, backup_thread_entry_frame, {
+            MVMuint32 backup_mark                   = MVM_gc_root_temp_mark(tc);
+            jmp_buf backup_interp_jump;
+            memcpy(backup_interp_jump, tc->interp_jump, sizeof(jmp_buf));
 
-        tc->cur_frame->return_value = &res;
-        tc->cur_frame->return_type  = MVM_RETURN_OBJ;
-        MVM_interp_run(tc, callback_invoke, &cid);
+            tc->cur_frame->return_value = &res;
+            tc->cur_frame->return_type  = MVM_RETURN_OBJ;
+            MVM_interp_run(tc, callback_invoke, &cid);
 
-        tc->interp_cur_op         = backup_interp_cur_op;
-        tc->interp_bytecode_start = backup_interp_bytecode_start;
-        tc->interp_reg_base       = backup_interp_reg_base;
-        tc->interp_cu             = backup_interp_cu;
-        tc->cur_frame             = backup_cur_frame;
-        tc->thread_entry_frame    = backup_thread_entry_frame;
-        memcpy(tc->interp_jump, backup_interp_jump, sizeof(jmp_buf));
-        MVM_gc_root_temp_mark_reset(tc, backup_mark);
+            tc->interp_cur_op         = backup_interp_cur_op;
+            tc->interp_bytecode_start = backup_interp_bytecode_start;
+            tc->interp_reg_base       = backup_interp_reg_base;
+            tc->interp_cu             = backup_interp_cu;
+            tc->cur_frame             = backup_cur_frame;
+            tc->current_frame_nr      = backup_cur_frame->sequence_nr;
+            tc->thread_entry_frame    = backup_thread_entry_frame;
+            memcpy(tc->interp_jump, backup_interp_jump, sizeof(jmp_buf));
+            MVM_gc_root_temp_mark_reset(tc, backup_mark);
+        });
+        });
     }
 
     /* Handle return value. */
@@ -493,6 +499,22 @@ MVMObject * MVM_nativecall_invoke(MVMThreadContext *tc, MVMObject *res_type,
                 values[i]           = MVM_malloc(sizeof(void *));
                 *(void **)values[i] = MVM_nativecall_unmarshal_cstruct(tc, value);
                 break;
+            case MVM_NATIVECALL_ARG_CPPSTRUCT: {
+                /* We need to allocate the struct (THIS) for C++ constructor before passing it along. */
+                if (i == 0 && !IS_CONCRETE(value)) {
+                    MVMCPPStructREPRData *repr_data = (MVMCPPStructREPRData *)STABLE(res_type)->REPR_data;
+                    /* Allocate a full byte aligned area where the C++ structure fits into. */
+                    void *ptr           = MVM_malloc(repr_data->struct_size > 0 ? repr_data->struct_size : 1);
+                    result              = MVM_nativecall_make_cppstruct(tc, res_type, ptr);
+                    values[i]           = MVM_malloc(sizeof(void *));
+                    *(void **)values[i] = ptr;
+                }
+                else {
+                    values[i]           = MVM_malloc(sizeof(void *));
+                    *(void **)values[i] = MVM_nativecall_unmarshal_cppstruct(tc, value);
+                }
+                break;
+            }
             case MVM_NATIVECALL_ARG_CPOINTER:
                 if ((arg_types[i] & MVM_NATIVECALL_ARG_RW_MASK) == MVM_NATIVECALL_ARG_RW) {
                     values[i]                     = MVM_malloc(sizeof(void *));
@@ -541,133 +563,148 @@ MVMObject * MVM_nativecall_invoke(MVMThreadContext *tc, MVMObject *res_type,
     }
 
 
-    /* Process return values. */
-    MVMROOT(tc, args, {
-    MVMROOT(tc, res_type, {
-        switch (ret_type & MVM_NATIVECALL_ARG_TYPE_MASK) {
-            case MVM_NATIVECALL_ARG_VOID:
-                ffi_call(&cif, entry_point, values[body->num_args], values);
-                result = res_type;
-                break;
-            case MVM_NATIVECALL_ARG_CHAR: {
-                signed char ret;
-                ffi_call(&cif, entry_point, &ret, values);
-                result = MVM_nativecall_make_int(tc, res_type, ret);
-                break;
+    if (result) {
+        /* We are calling a C++ constructor so we hand back the invocant (THIS) we recorded earlier. */
+        void *ret; // We are not going to use it, but we need to pass it to libffi.
+        ffi_call(&cif, entry_point, &ret, values);
+    }
+    else {
+        /* Process return values. */
+        MVMROOT(tc, args, {
+        MVMROOT(tc, res_type, {
+            switch (ret_type & MVM_NATIVECALL_ARG_TYPE_MASK) {
+                case MVM_NATIVECALL_ARG_VOID: {
+                    void *ret;
+                    ffi_call(&cif, entry_point, &ret, values);
+                    result = res_type;
+                    break;
+                }
+                case MVM_NATIVECALL_ARG_CHAR: {
+                    signed char ret;
+                    ffi_call(&cif, entry_point, &ret, values);
+                    result = MVM_nativecall_make_int(tc, res_type, ret);
+                    break;
+                }
+                case MVM_NATIVECALL_ARG_SHORT: {
+                    signed short ret;
+                    ffi_call(&cif, entry_point, &ret, values);
+                    result = MVM_nativecall_make_int(tc, res_type, ret);
+                    break;
+                }
+                case MVM_NATIVECALL_ARG_INT: {
+                    signed int ret;
+                    ffi_call(&cif, entry_point, &ret, values);
+                    result = MVM_nativecall_make_int(tc, res_type, ret);
+                    break;
+                }
+                case MVM_NATIVECALL_ARG_LONG: {
+                    signed long ret;
+                    ffi_call(&cif, entry_point, &ret, values);
+                    result = MVM_nativecall_make_int(tc, res_type, ret);
+                    break;
+                }
+                case MVM_NATIVECALL_ARG_LONGLONG: {
+                    signed long long ret;
+                    ffi_call(&cif, entry_point, &ret, values);
+                    result = MVM_nativecall_make_int(tc, res_type, ret);
+                    break;
+                }
+                case MVM_NATIVECALL_ARG_FLOAT: {
+                    float ret;
+                    ffi_call(&cif, entry_point, &ret, values);
+                    result = MVM_nativecall_make_num(tc, res_type, ret);
+                    break;
+                }
+                case MVM_NATIVECALL_ARG_DOUBLE: {
+                    double ret;
+                    ffi_call(&cif, entry_point, &ret, values);
+                    result = MVM_nativecall_make_num(tc, res_type, ret);
+                    break;
+                }
+                case MVM_NATIVECALL_ARG_ASCIISTR:
+                case MVM_NATIVECALL_ARG_UTF8STR:
+                case MVM_NATIVECALL_ARG_UTF16STR: {
+                    char *ret;
+                    ffi_call(&cif, entry_point, &ret, values);
+                    result = MVM_nativecall_make_str(tc, res_type, body->ret_type, ret);
+                    break;
+                }
+                case MVM_NATIVECALL_ARG_CSTRUCT: {
+                    void *ret;
+                    ffi_call(&cif, entry_point, &ret, values);
+                    result = MVM_nativecall_make_cstruct(tc, res_type, ret);
+                    break;
+                }
+                case MVM_NATIVECALL_ARG_CPPSTRUCT: {
+                    void *ret;
+                    ffi_call(&cif, entry_point, &ret, values);
+                    result = MVM_nativecall_make_cppstruct(tc, res_type, ret);
+                    break;
+                }
+                case MVM_NATIVECALL_ARG_CPOINTER: {
+                    void *ret;
+                    ffi_call(&cif, entry_point, &ret, values);
+                    result = MVM_nativecall_make_cpointer(tc, res_type, ret);
+                    break;
+                }
+                case MVM_NATIVECALL_ARG_CARRAY: {
+                    void *ret;
+                    ffi_call(&cif, entry_point, &ret, values);
+                    result = MVM_nativecall_make_carray(tc, res_type, ret);
+                    break;
+                }
+                case MVM_NATIVECALL_ARG_CUNION: {
+                    void *ret;
+                    ffi_call(&cif, entry_point, &ret, values);
+                    result = MVM_nativecall_make_cunion(tc, res_type, ret);
+                    break;
+                }
+                case MVM_NATIVECALL_ARG_CALLBACK: {
+                    /* TODO: A callback -return- value means that we have a C method
+                    * that needs to be wrapped similarly to a is native(...) Perl 6
+                    * sub. */
+                    void *ret;
+                    ffi_call(&cif, entry_point, &ret, values);
+                    /* XXX do something with the function pointer: ret */
+                    result = res_type;
+                    break;
+                }
+                case MVM_NATIVECALL_ARG_UCHAR: {
+                    unsigned char ret;
+                    ffi_call(&cif, entry_point, &ret, values);
+                    result = MVM_nativecall_make_int(tc, res_type, ret);
+                    break;
+                }
+                case MVM_NATIVECALL_ARG_USHORT: {
+                    unsigned short ret;
+                    ffi_call(&cif, entry_point, &ret, values);
+                    result = MVM_nativecall_make_int(tc, res_type, ret);
+                    break;
+                }
+                case MVM_NATIVECALL_ARG_UINT: {
+                    unsigned int ret;
+                    ffi_call(&cif, entry_point, &ret, values);
+                    result = MVM_nativecall_make_int(tc, res_type, ret);
+                    break;
+                }
+                case MVM_NATIVECALL_ARG_ULONG: {
+                    unsigned long ret;
+                    ffi_call(&cif, entry_point, &ret, values);
+                    result = MVM_nativecall_make_int(tc, res_type, ret);
+                    break;
+                }
+                case MVM_NATIVECALL_ARG_ULONGLONG: {
+                    unsigned long long ret;
+                    ffi_call(&cif, entry_point, &ret, values);
+                    result = MVM_nativecall_make_int(tc, res_type, ret);
+                    break;
+                }
+                default:
+                    MVM_exception_throw_adhoc(tc, "Internal error: unhandled libffi return type");
             }
-            case MVM_NATIVECALL_ARG_SHORT: {
-                signed short ret;
-                ffi_call(&cif, entry_point, &ret, values);
-                result = MVM_nativecall_make_int(tc, res_type, ret);
-                break;
-            }
-            case MVM_NATIVECALL_ARG_INT: {
-                signed int ret;
-                ffi_call(&cif, entry_point, &ret, values);
-                result = MVM_nativecall_make_int(tc, res_type, ret);
-                break;
-            }
-            case MVM_NATIVECALL_ARG_LONG: {
-                signed long ret;
-                ffi_call(&cif, entry_point, &ret, values);
-                result = MVM_nativecall_make_int(tc, res_type, ret);
-                break;
-            }
-            case MVM_NATIVECALL_ARG_LONGLONG: {
-                signed long long ret;
-                ffi_call(&cif, entry_point, &ret, values);
-                result = MVM_nativecall_make_int(tc, res_type, ret);
-                break;
-            }
-            case MVM_NATIVECALL_ARG_FLOAT: {
-                float ret;
-                ffi_call(&cif, entry_point, &ret, values);
-                result = MVM_nativecall_make_num(tc, res_type, ret);
-                break;
-            }
-            case MVM_NATIVECALL_ARG_DOUBLE: {
-                double ret;
-                ffi_call(&cif, entry_point, &ret, values);
-                result = MVM_nativecall_make_num(tc, res_type, ret);
-                break;
-            }
-            case MVM_NATIVECALL_ARG_ASCIISTR:
-            case MVM_NATIVECALL_ARG_UTF8STR:
-            case MVM_NATIVECALL_ARG_UTF16STR: {
-                char *ret;
-                ffi_call(&cif, entry_point, &ret, values);
-                result = MVM_nativecall_make_str(tc, res_type, body->ret_type, ret);
-                break;
-            }
-            case MVM_NATIVECALL_ARG_CSTRUCT: {
-                void *ret;
-                ffi_call(&cif, entry_point, &ret, values);
-                result = MVM_nativecall_make_cstruct(tc, res_type, ret);
-                break;
-            }
-            case MVM_NATIVECALL_ARG_CPOINTER: {
-                void *ret;
-                ffi_call(&cif, entry_point, &ret, values);
-                result = MVM_nativecall_make_cpointer(tc, res_type, ret);
-                break;
-            }
-            case MVM_NATIVECALL_ARG_CARRAY: {
-                void *ret;
-                ffi_call(&cif, entry_point, &ret, values);
-                result = MVM_nativecall_make_carray(tc, res_type, ret);
-                break;
-            }
-            case MVM_NATIVECALL_ARG_CUNION: {
-                void *ret;
-                ffi_call(&cif, entry_point, &ret, values);
-                result = MVM_nativecall_make_cunion(tc, res_type, ret);
-                break;
-            }
-            case MVM_NATIVECALL_ARG_CALLBACK: {
-                /* TODO: A callback -return- value means that we have a C method
-                * that needs to be wrapped similarly to a is native(...) Perl 6
-                * sub. */
-                void *ret;
-                ffi_call(&cif, entry_point, &ret, values);
-                /* XXX do something with the function pointer: ret */
-                result = res_type;
-                break;
-            }
-            case MVM_NATIVECALL_ARG_UCHAR: {
-                unsigned char ret;
-                ffi_call(&cif, entry_point, &ret, values);
-                result = MVM_nativecall_make_int(tc, res_type, ret);
-                break;
-            }
-            case MVM_NATIVECALL_ARG_USHORT: {
-                unsigned short ret;
-                ffi_call(&cif, entry_point, &ret, values);
-                result = MVM_nativecall_make_int(tc, res_type, ret);
-                break;
-            }
-            case MVM_NATIVECALL_ARG_UINT: {
-                unsigned int ret;
-                ffi_call(&cif, entry_point, &ret, values);
-                result = MVM_nativecall_make_int(tc, res_type, ret);
-                break;
-            }
-            case MVM_NATIVECALL_ARG_ULONG: {
-                unsigned long ret;
-                ffi_call(&cif, entry_point, &ret, values);
-                result = MVM_nativecall_make_int(tc, res_type, ret);
-                break;
-            }
-            case MVM_NATIVECALL_ARG_ULONGLONG: {
-                unsigned long long ret;
-                ffi_call(&cif, entry_point, &ret, values);
-                result = MVM_nativecall_make_int(tc, res_type, ret);
-                break;
-            }
-            default:
-                MVM_exception_throw_adhoc(tc, "Internal error: unhandled libffi return type");
-        }
-    });
-    });
+        });
+        });
+    }
 
     for (i = 0; i < num_args; i++) {
         MVMObject *value = MVM_repr_at_pos_o(tc, args, i);
