@@ -24,7 +24,6 @@ static void add_work(MVMThreadContext *tc, MVMThreadContext *stolen) {
  * the run, and are not counted. Returns the count of threads that should be
  * added to the finished countdown. */
 static MVMuint32 signal_one_thread(MVMThreadContext *tc, MVMThreadContext *to_signal) {
-
     /* Loop here since we may not succeed first time (e.g. the status of the
      * thread may change between the two ways we try to twiddle it). */
     while (1) {
@@ -156,23 +155,36 @@ static void finish_gc(MVMThreadContext *tc, MVMuint8 gen, MVMuint8 is_coordinato
      * in-trays are settled, coordinator walks threads looking for anything
      * that needs adding to the finalize queue. It then will make another
      * iteration over in-trays to handle cross-thread references to objects
-     * needing finalization. */
+     * needing finalization. For full collections, collected objects are then
+     * cleaned from all inter-generational sets, and finally any objects to
+     * be freed at the fixed size allocator's next safepoint are freed. */
     if (is_coordinator) {
         GCDEBUG_LOG(tc, MVM_GC_DEBUG_ORCHESTRATE,
             "Thread %d run %d : Co-ordinator handling in-tray clearing completion\n");
         clear_intrays(tc, gen);
 
+        GCDEBUG_LOG(tc, MVM_GC_DEBUG_ORCHESTRATE,
+            "Thread %d run %d : Co-ordinator handling finalizers\n");
         MVM_finalize_walk_queues(tc, gen);
         clear_intrays(tc, gen);
 
         if (gen == MVMGCGenerations_Both) {
             MVMThread *cur_thread = (MVMThread *)MVM_load(&tc->instance->threads);
+            GCDEBUG_LOG(tc, MVM_GC_DEBUG_ORCHESTRATE,
+                "Thread %d run %d : Co-ordinator handling inter-gen root cleanup\n");
             while (cur_thread) {
                 if (cur_thread->body.tc)
                     MVM_gc_root_gen2_cleanup(cur_thread->body.tc);
                 cur_thread = cur_thread->body.next;
             }
         }
+
+        GCDEBUG_LOG(tc, MVM_GC_DEBUG_ORCHESTRATE,
+            "Thread %d run %d : Co-ordinator handling fixed-size allocator safepoint frees\n");
+        MVM_fixed_size_safepoint(tc, tc->instance->fsa);
+
+        MVM_profile_heap_take_snapshot(tc);
+
         GCDEBUG_LOG(tc, MVM_GC_DEBUG_ORCHESTRATE,
             "Thread %d run %d : Co-ordinator signalling in-trays clear\n");
         MVM_store(&tc->instance->gc_intrays_clearing, 0);
@@ -241,7 +253,8 @@ void MVM_gc_mark_thread_blocked(MVMThreadContext *tc) {
         if (MVM_load(&tc->gc_status) == MVMGCStatus_INTERRUPT)
             MVM_gc_enter_from_interrupt(tc);
         else
-            MVM_panic(MVM_exitcode_gcorch, "Invalid GC status observed; aborting");
+            MVM_panic(MVM_exitcode_gcorch,
+                "Invalid GC status observed while blocking thread; aborting");
     }
 }
 
@@ -259,9 +272,19 @@ void MVM_gc_mark_thread_unblocked(MVMThreadContext *tc) {
 }
 
 static MVMint32 is_full_collection(MVMThreadContext *tc) {
-    MVMuint64 threshold = MVM_GC_GEN2_THRESHOLD_BASE +
-        (tc->instance->num_user_threads * MVM_GC_GEN2_THRESHOLD_THREAD);
-    return MVM_load(&tc->instance->gc_promoted_bytes_since_last_full) > threshold;
+    MVMuint64 percent_growth;
+    size_t rss, promoted;
+
+    /* If it's below the absolute minimum, quickly return. */
+    promoted = (size_t)MVM_load(&tc->instance->gc_promoted_bytes_since_last_full);
+    if (promoted < MVM_GC_GEN2_THRESHOLD_MINIMUM)
+        return 0;
+
+    /* Otherwise, consider percentage of resident set size. */
+    if (uv_resident_set_memory(&rss) < 0 || rss == 0)
+        rss = 50 * 1024 * 1024;
+    percent_growth = (100 * promoted) / rss;
+    return percent_growth >= MVM_GC_GEN2_THRESHOLD_PERCENT;
 }
 
 static void run_gc(MVMThreadContext *tc, MVMuint8 what_to_do) {
@@ -269,7 +292,7 @@ static void run_gc(MVMThreadContext *tc, MVMuint8 what_to_do) {
     MVMuint32  i, n;
 
     /* Decide nursery or full collection. */
-    gen = is_full_collection(tc) ? MVMGCGenerations_Both : MVMGCGenerations_Nursery;
+    gen = tc->instance->gc_full_collect ? MVMGCGenerations_Both : MVMGCGenerations_Nursery;
 
     /* Do GC work for ourselves and any work threads. */
     for (i = 0, n = tc->gc_work_count ; i < n; i++) {
@@ -307,7 +330,7 @@ static void run_gc(MVMThreadContext *tc, MVMuint8 what_to_do) {
             GCDEBUG_LOG(tc, MVM_GC_DEBUG_ORCHESTRATE,
                 "Thread %d run %d : freeing gen2 of thread %d\n",
                 other->thread_id);
-            MVM_gc_collect_free_gen2_unmarked(other);
+            MVM_gc_collect_free_gen2_unmarked(other, 0);
         }
     }
 }
@@ -323,7 +346,6 @@ void MVM_gc_enter_from_allocator(MVMThreadContext *tc) {
     if (MVM_trycas(&tc->instance->gc_start, 0, 1)) {
         MVMThread *last_starter = NULL;
         MVMuint32 num_threads = 0;
-        MVMuint32 is_full;
 
         /* Need to wait for other threads to reset their gc_status. */
         while (MVM_load(&tc->instance->gc_ack)) {
@@ -341,11 +363,11 @@ void MVM_gc_enter_from_allocator(MVMThreadContext *tc) {
             (int)MVM_load(&tc->instance->gc_seq_number));
 
         /* Decide if it will be a full collection. */
-        is_full = is_full_collection(tc);
+        tc->instance->gc_full_collect = is_full_collection(tc);
 
         /* If profiling, record that GC is starting. */
         if (tc->instance->profiling)
-            MVM_profiler_log_gc_start(tc, is_full);
+            MVM_profiler_log_gc_start(tc, tc->instance->gc_full_collect);
 
         /* Ensure our stolen list is empty. */
         tc->gc_work_count = 0;
@@ -396,9 +418,16 @@ void MVM_gc_enter_from_allocator(MVMThreadContext *tc) {
 
         /* Now we're ready to start, zero promoted since last full collection
          * counter if this is a full collect. */
-        if (is_full)
+        if (tc->instance->gc_full_collect)
             MVM_store(&tc->instance->gc_promoted_bytes_since_last_full, 0);
 
+        /* This is a safe point for us to free any STables that have been marked
+         * for deletion in the previous collection (since we let finalization -
+         * which appends to this list - happen after we set threads on their
+         * way again, it's not safe to do it in the previous collection). */
+        GCDEBUG_LOG(tc, MVM_GC_DEBUG_ORCHESTRATE, "Thread %d run %d : Freeing STables if needed\n");
+        MVM_gc_collect_free_stables(tc);
+
         /* Signal to the rest to start */
         GCDEBUG_LOG(tc, MVM_GC_DEBUG_ORCHESTRATE, "Thread %d run %d : coordinator signalling start\n");
         if (MVM_decr(&tc->instance->gc_start) != 1)
@@ -408,12 +437,6 @@ void MVM_gc_enter_from_allocator(MVMThreadContext *tc) {
         GCDEBUG_LOG(tc, MVM_GC_DEBUG_ORCHESTRATE, "Thread %d run %d : coordinator entering run_gc\n");
         run_gc(tc, MVMGCWhatToDo_All);
 
-        /* Free any STables that have been marked for deletion. It's okay for
-         * us to muck around in another thread's fromspace while it's mutating
-         * tospace, really. */
-        GCDEBUG_LOG(tc, MVM_GC_DEBUG_ORCHESTRATE, "Thread %d run %d : Freeing STables if needed\n");
-        MVM_gc_collect_free_stables(tc);
-
         /* If profiling, record that GC is over. */
         if (tc->instance->profiling)
             MVM_profiler_log_gc_end(tc);
@@ -487,6 +510,6 @@ void MVM_gc_global_destruction(MVMThreadContext *tc) {
     /* Run the objects' finalizers */
     MVM_gc_collect_free_nursery_uncopied(tc, tc->nursery_alloc);
     MVM_gc_root_gen2_cleanup(tc);
-    MVM_gc_collect_free_gen2_unmarked(tc);
+    MVM_gc_collect_free_gen2_unmarked(tc, 1);
     MVM_gc_collect_free_stables(tc);
 }
