@@ -176,8 +176,8 @@ MVMint64 MVM_proc_shell(MVMThreadContext *tc, MVMString *cmd, MVMString *cwd, MV
     char * const cmdin = MVM_string_utf8_c8_encode_C_string(tc, cmd);
     char * const _cwd = MVM_string_utf8_c8_encode_C_string(tc, cwd);
     const MVMuint64 size = MVM_repr_elems(tc, env);
-    MVMIter * const iter = (MVMIter *)MVM_iter(tc, env);
     char **_env = MVM_malloc((size + 1) * sizeof(char *));
+    MVMIter *iter;
 
 #ifdef _WIN32
     const MVMuint16 acp = GetACP(); /* We should get ACP at runtime. */
@@ -195,10 +195,19 @@ MVMint64 MVM_proc_shell(MVMThreadContext *tc, MVMString *cmd, MVMString *cwd, MV
     args[3] = NULL;
 #endif
 
-    INIT_ENV();
+    MVMROOT(tc, in, {
+    MVMROOT(tc, out, {
+    MVMROOT(tc, err, {
+        iter = (MVMIter *)MVM_iter(tc, env);
+        INIT_ENV();
+    });
+    });
+    });
+
     setup_process_stdio(tc, in,  process, &process_stdio[0], 0, flags,      "shell");
     setup_process_stdio(tc, out, process, &process_stdio[1], 1, flags >> 3, "shell");
-    setup_process_stdio(tc, err, process, &process_stdio[2], 2, flags >> 6, "shell");
+    if (!(flags & MVM_PIPE_MERGED_OUT_ERR))
+        setup_process_stdio(tc, err, process, &process_stdio[2], 2, flags >> 6, "shell");
 
     process_options.stdio       = process_stdio;
     process_options.file        = _cmd;
@@ -206,7 +215,11 @@ MVMint64 MVM_proc_shell(MVMThreadContext *tc, MVMString *cmd, MVMString *cwd, MV
     process_options.cwd         = _cwd;
     process_options.flags       = UV_PROCESS_WINDOWS_VERBATIM_ARGUMENTS | UV_PROCESS_WINDOWS_HIDE;
     process_options.env         = _env;
-    process_options.stdio_count = 3;
+    if (flags & MVM_PIPE_MERGED_OUT_ERR) {
+        process_options.stdio_count = 2;
+    }
+    else
+        process_options.stdio_count = 3;
     process_options.exit_cb     = spawn_on_exit;
     if (flags & (MVM_PIPE_CAPTURE_IN | MVM_PIPE_CAPTURE_OUT | MVM_PIPE_CAPTURE_ERR)) {
         process_still_running = 1;
@@ -251,11 +264,11 @@ MVMint64 MVM_proc_spawn(MVMThreadContext *tc, MVMObject *argv, MVMString *cwd, M
 
     char   * const      _cwd = MVM_string_utf8_c8_encode_C_string(tc, cwd);
     const MVMuint64     size = MVM_repr_elems(tc, env);
-    MVMIter * const     iter = (MVMIter *)MVM_iter(tc, env);
     char              **_env = MVM_malloc((size + 1) * sizeof(char *));
     const MVMuint64  arg_size = MVM_repr_elems(tc, argv);
     char             **args = MVM_malloc((arg_size + 1) * sizeof(char *));
     MVMRegister        reg;
+    MVMIter           *iter;
 
     i = 0;
     while(i < arg_size) {
@@ -264,10 +277,19 @@ MVMint64 MVM_proc_spawn(MVMThreadContext *tc, MVMObject *argv, MVMString *cwd, M
     }
     args[arg_size] = NULL;
 
-    INIT_ENV();
+    MVMROOT(tc, in, {
+    MVMROOT(tc, out, {
+    MVMROOT(tc, err, {
+        iter = (MVMIter *)MVM_iter(tc, env);
+        INIT_ENV();
+    });
+    });
+    });
+
     setup_process_stdio(tc, in,  process, &process_stdio[0], 0, flags,      "spawn");
     setup_process_stdio(tc, out, process, &process_stdio[1], 1, flags >> 3, "spawn");
-    setup_process_stdio(tc, err, process, &process_stdio[2], 2, flags >> 6, "spawn");
+    if (!(flags & MVM_PIPE_MERGED_OUT_ERR))
+        setup_process_stdio(tc, err, process, &process_stdio[2], 2, flags >> 6, "spawn");
 
     process_options.stdio       = process_stdio;
     process_options.file        = arg_size ? args[0] : NULL;
@@ -275,7 +297,11 @@ MVMint64 MVM_proc_spawn(MVMThreadContext *tc, MVMObject *argv, MVMString *cwd, M
     process_options.cwd         = _cwd;
     process_options.flags       = UV_PROCESS_WINDOWS_HIDE;
     process_options.env         = _env;
-    process_options.stdio_count = 3;
+    if (flags & MVM_PIPE_MERGED_OUT_ERR) {
+        process_options.stdio_count = 2;
+    }
+    else
+        process_options.stdio_count = 3;
     process_options.exit_cb     = spawn_on_exit;
     if (flags & (MVM_PIPE_CAPTURE_IN | MVM_PIPE_CAPTURE_OUT | MVM_PIPE_CAPTURE_ERR)) {
         process->data = MVM_calloc(1, sizeof(MVMint64));
@@ -1093,7 +1119,7 @@ void MVM_proc_kill_async(MVMThreadContext *tc, MVMObject *handle_obj, MVMint64 s
             /* It's fine; send the kill by cancelling the task. */
             MVMIOAsyncProcessData *data = (MVMIOAsyncProcessData *)handle->body.data;
             data->signal = signal;
-            MVM_io_eventloop_cancel_work(tc, data->async_task);
+            MVM_io_eventloop_cancel_work(tc, data->async_task, NULL, NULL);
             return;
         }
     }
