@@ -83,13 +83,13 @@ static void from_num(MVMnum64 d, mp_int *a) {
     lower = rest / d_digit;
     lowest = fmod(rest,d_digit );
     if (upper >= 1) {
-        mp_set_long(a, (unsigned long) upper);
+        MVM_bigint_mp_set_uint64(a, (MVMuint64) upper);
         mp_mul_2d(a, DIGIT_BIT , a);
         DIGIT(a, 0) = (mp_digit) lower;
         mp_mul_2d(a, DIGIT_BIT , a);
     } else {
         if (lower >= 1) {
-            mp_set_long(a, (unsigned long) lower);
+            MVM_bigint_mp_set_uint64(a, (MVMuint64) lower);
             mp_mul_2d(a, DIGIT_BIT , a);
             a->used = 2;
         } else {
@@ -845,12 +845,14 @@ MVMnum64 MVM_bigint_to_num(MVMThreadContext *tc, MVMObject *a) {
     }
 }
 
-void MVM_bigint_from_num(MVMThreadContext *tc, MVMObject *a, MVMnum64 n) {
-    MVMP6bigintBody *ba = get_bigint_body(tc, a);
+MVMObject *MVM_bigint_from_num(MVMThreadContext *tc, MVMObject *result_type, MVMnum64 n) {
+    MVMObject * const result = MVM_repr_alloc_init(tc, result_type);
+    MVMP6bigintBody *ba = get_bigint_body(tc, result);
     mp_int *ia = MVM_malloc(sizeof(mp_int));
     mp_init(ia);
     from_num(n, ia);
     store_bigint_result(ba, ia);
+    return result;
 }
 
 MVMnum64 MVM_bigint_div_num(MVMThreadContext *tc, MVMObject *a, MVMObject *b) {
