@@ -132,8 +132,9 @@ MVMInstance * MVM_vm_create_instance(void) {
     /* Bootstrap 6model. It is assumed the GC will not be called during this. */
     MVM_6model_bootstrap(instance->main_thread);
 
-    /* Fix up main thread's usecapture. */
+    /* Fix up main thread's usecapture and last_payload. */
     instance->main_thread->cur_usecapture = MVM_repr_alloc_init(instance->main_thread, instance->CallCapture);
+    instance->main_thread->last_payload = instance->VMNull;
 
     /* Initialize event loop thread starting mutex. */
     init_mutex(instance->mutex_event_loop_start, "event loop thread start");
@@ -397,6 +398,7 @@ void MVM_vm_destroy_instance(MVMInstance *instance) {
     /* Clean up GC permanent roots related resources. */
     uv_mutex_destroy(&instance->mutex_permroots);
     MVM_free(instance->permroots);
+    MVM_free(instance->permroot_descriptions);
 
     /* Clean up Hash of HLLConfig. */
     uv_mutex_destroy(&instance->mutex_hllconfigs);
@@ -441,9 +443,6 @@ void MVM_vm_destroy_instance(MVMInstance *instance) {
     uv_mutex_destroy(&instance->mutex_callsite_interns);
     cleanup_callsite_interns(instance);
 
-    /* Clean up fixed size allocator */
-    MVM_fixed_size_destroy(instance->fsa);
-
     /* Release this interpreter's hold on Unicode database */
     MVM_unicode_release(instance->main_thread);
 
@@ -453,11 +452,19 @@ void MVM_vm_destroy_instance(MVMInstance *instance) {
         fclose(instance->spesh_log_fh);
     if (instance->jit_log_fh)
         fclose(instance->jit_log_fh);
+    if (instance->dynvar_log_fh)
+        fclose(instance->dynvar_log_fh);
+
+    /* Clean up cross-thread-write-logging mutex */
+    uv_mutex_destroy(&instance->mutex_cross_thread_write_logging);
 
     /* Clean up NFG. */
     uv_mutex_destroy(&instance->nfg->update_mutex);
     MVM_nfg_destroy(instance->main_thread);
 
+    /* Clean up fixed size allocator */
+    MVM_fixed_size_destroy(instance->fsa);
+
     /* Clean up integer constant and string cache. */
     uv_mutex_destroy(&instance->mutex_int_const_cache);
     MVM_free(instance->int_const_cache);
