@@ -18,8 +18,8 @@ static void copy_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMuint16 to_orig
 
 /* Called when one set of facts depend on another, allowing any log guard
  * that is to thank to be marked used as needed later on. */
-static void depend(MVMThreadContext *tc, MVMSpeshGraph *g,
-                   MVMSpeshFacts *target, MVMSpeshFacts *source) {
+void MVM_spesh_facts_depend(MVMThreadContext *tc, MVMSpeshGraph *g,
+                            MVMSpeshFacts *target, MVMSpeshFacts *source) {
     if (source->flags & MVM_SPESH_FACT_FROM_LOG_GUARD) {
         target->flags     |= MVM_SPESH_FACT_FROM_LOG_GUARD;
         target->log_guard  = source->log_guard;
@@ -36,7 +36,7 @@ static void create_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMuint16 obj_o
     if (type_facts->flags & MVM_SPESH_FACT_KNOWN_TYPE) {
         obj_facts->type   = type_facts->type;
         obj_facts->flags |= MVM_SPESH_FACT_KNOWN_TYPE;
-        depend(tc, g, obj_facts, type_facts);
+        MVM_spesh_facts_depend(tc, g, obj_facts, type_facts);
     }
 
     /* We know it's a concrete object. */
@@ -95,9 +95,48 @@ static void object_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMuint16 tgt_o
     }
 }
 
+/* Checks if there's a possible aliasing operation that could cause the
+ * facts about the contents of a container to be invalid by the instruction
+ * under consideration. Assumes the instruction is a decont with argument 1
+ * being the thing to decontainerize. */
+MVMint32 MVM_spesh_facts_decont_blocked_by_alias(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshIns *ins) {
+    /* No facts or no writer means we can't know it's safe; block. */
+    MVMSpeshFacts *facts = MVM_spesh_get_facts(tc, g, ins->operands[1]);
+    if (!facts || !facts->writer)
+        return 1;
+
+    /* Walk backwards over instructions. */
+    while (ins->prev) {
+        ins = ins->prev;
+
+        /* If we found the writer without anything blocking, we're good. */
+        if (ins == facts->writer)
+            return 0;
+
+        /* If there's an operation that may alias, blocked. */
+        switch (ins->info->opcode) {
+            case MVM_OP_bindattr_o:
+            case MVM_OP_bindattrs_o:
+            case MVM_OP_assign:
+            case MVM_OP_assignunchecked:
+            case MVM_OP_assign_i:
+            case MVM_OP_assign_n:
+            case MVM_OP_assign_s:
+            case MVM_OP_sp_bind_o:
+            case MVM_OP_sp_p6obind_o:
+                return 1;
+        }
+    }
+
+    /* We didn't find the writer in this basic block, meaning an invocation
+     * may have made it unsafe. Block. */
+    return 1;
+}
+
 /* Propagates information relating to decontainerization. */
-static void decont_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMuint16 out_orig,
-                         MVMuint16 out_i, MVMuint16 in_orig, MVMuint16 in_i) {
+static void decont_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshIns *ins,
+                         MVMuint16 out_orig, MVMuint16 out_i, MVMuint16 in_orig,
+                         MVMuint16 in_i) {
     MVMSpeshFacts *out_facts = &(g->facts[out_orig][out_i]);
     MVMSpeshFacts *in_facts  = &(g->facts[in_orig][in_i]);
 
@@ -112,18 +151,20 @@ static void decont_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMuint16 out_o
 
     /* We may also know the original was containerized, and have some facts
      * about its contents. */
-    if (in_flags & MVM_SPESH_FACT_KNOWN_DECONT_TYPE) {
-        out_facts->type = in_facts->decont_type;
-        out_facts->flags |= MVM_SPESH_FACT_KNOWN_TYPE;
+    if (!MVM_spesh_facts_decont_blocked_by_alias(tc, g, ins)) {
+        if (in_flags & MVM_SPESH_FACT_KNOWN_DECONT_TYPE) {
+            out_facts->type = in_facts->decont_type;
+            out_facts->flags |= MVM_SPESH_FACT_KNOWN_TYPE;
+        }
+        if (in_flags & MVM_SPESH_FACT_DECONT_CONCRETE)
+            out_facts->flags |= MVM_SPESH_FACT_CONCRETE;
+        else if (in_flags & MVM_SPESH_FACT_DECONT_TYPEOBJ)
+            out_facts->flags |= MVM_SPESH_FACT_TYPEOBJ;
+        if (in_flags & (MVM_SPESH_FACT_KNOWN_DECONT_TYPE |
+                        MVM_SPESH_FACT_DECONT_CONCRETE |
+                        MVM_SPESH_FACT_DECONT_TYPEOBJ))
+            MVM_spesh_facts_depend(tc, g, out_facts, in_facts);
     }
-    if (in_flags & MVM_SPESH_FACT_DECONT_CONCRETE)
-        out_facts->flags |= MVM_SPESH_FACT_CONCRETE;
-    else if (in_flags & MVM_SPESH_FACT_DECONT_TYPEOBJ)
-        out_facts->flags |= MVM_SPESH_FACT_TYPEOBJ;
-    if (in_flags & (MVM_SPESH_FACT_KNOWN_DECONT_TYPE |
-                    MVM_SPESH_FACT_DECONT_CONCRETE |
-                    MVM_SPESH_FACT_DECONT_TYPEOBJ))
-        depend(tc, g, out_facts, in_facts);
 }
 
 /* Looks up a wval and adds information based on it. */
@@ -503,7 +544,7 @@ static void add_bb_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb,
                 g->sf->body.cu->body.hll_config->slurpy_hash_type);
             break;
         case MVM_OP_decont:
-            decont_facts(tc, g,
+            decont_facts(tc, g, ins,
                 ins->operands[0].reg.orig, ins->operands[0].reg.i,
                 ins->operands[1].reg.orig, ins->operands[1].reg.i);
             break;
@@ -544,6 +585,7 @@ static void add_bb_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb,
             create_facts_with_type(tc, g,
                 ins->operands[0].reg.orig, ins->operands[0].reg.i,
                 g->sf->body.cu->body.hll_config->int_lex_ref);
+            break;
         case MVM_OP_getregref_n:
         case MVM_OP_getlexref_n:
         case MVM_OP_getregref_n32:
@@ -551,38 +593,46 @@ static void add_bb_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb,
             create_facts_with_type(tc, g,
                 ins->operands[0].reg.orig, ins->operands[0].reg.i,
                 g->sf->body.cu->body.hll_config->num_lex_ref);
+            break;
         case MVM_OP_getregref_s:
         case MVM_OP_getlexref_s:
             create_facts_with_type(tc, g,
                 ins->operands[0].reg.orig, ins->operands[0].reg.i,
                 g->sf->body.cu->body.hll_config->str_lex_ref);
+            break;
         case MVM_OP_getattrref_i:
         case MVM_OP_getattrsref_i:
             create_facts_with_type(tc, g,
                 ins->operands[0].reg.orig, ins->operands[0].reg.i,
                 g->sf->body.cu->body.hll_config->int_attr_ref);
+            break;
         case MVM_OP_getattrref_n:
         case MVM_OP_getattrsref_n:
             create_facts_with_type(tc, g,
                 ins->operands[0].reg.orig, ins->operands[0].reg.i,
                 g->sf->body.cu->body.hll_config->num_attr_ref);
+            break;
         case MVM_OP_getattrref_s:
         case MVM_OP_getattrsref_s:
             create_facts_with_type(tc, g,
                 ins->operands[0].reg.orig, ins->operands[0].reg.i,
                 g->sf->body.cu->body.hll_config->str_attr_ref);
+            break;
         case MVM_OP_atposref_i:
             create_facts_with_type(tc, g,
                 ins->operands[0].reg.orig, ins->operands[0].reg.i,
                 g->sf->body.cu->body.hll_config->int_pos_ref);
+            break;
         case MVM_OP_atposref_n:
             create_facts_with_type(tc, g,
                 ins->operands[0].reg.orig, ins->operands[0].reg.i,
                 g->sf->body.cu->body.hll_config->num_pos_ref);
+            break;
         case MVM_OP_atposref_s:
             create_facts_with_type(tc, g,
                 ins->operands[0].reg.orig, ins->operands[0].reg.i,
                 g->sf->body.cu->body.hll_config->str_pos_ref);
+            break;
 
         case MVM_OP_const_i64:
         case MVM_OP_const_i32:
