@@ -228,6 +228,8 @@ static void optimize_isconcrete(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpesh
         ins->operands[1].lit_i16    = result_facts->value.i;
 
         MVM_spesh_use_facts(tc, g, obj_facts);
+        MVM_spesh_facts_depend(tc, g, result_facts, obj_facts);
+
         obj_facts->usages--;
     }
 }
@@ -245,7 +247,9 @@ static void optimize_exception_ops(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSp
         ins->operands[1].lit_i16 = st->size;
         ins->operands[2].lit_i16 = MVM_spesh_add_spesh_slot(tc, g, (MVMCollectable *)st);
     } else {
+        /*
         MVMSpeshFacts *target_facts;
+        */
 
         /* XXX This currently still causes problems. */
         return;
@@ -597,26 +601,34 @@ static void optimize_decont(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *
         copy_facts(tc, g, ins->operands[0], ins->operands[1]);
     }
     else {
-        MVMSpeshFacts *res_facts;
         if (obj_facts->flags & MVM_SPESH_FACT_KNOWN_TYPE && obj_facts->type) {
             MVMSTable *stable = STABLE(obj_facts->type);
             MVMContainerSpec const *contspec = stable->container_spec;
             if (contspec && contspec->fetch_never_invokes && contspec->spesh) {
                 contspec->spesh(tc, stable, g, bb, ins);
+                MVM_spesh_use_facts(tc, g, obj_facts);
             }
         }
 
-        MVM_spesh_use_facts(tc, g, obj_facts);
-
-        res_facts = MVM_spesh_get_facts(tc, g, ins->operands[0]);
-        if (obj_facts->flags & MVM_SPESH_FACT_KNOWN_DECONT_TYPE) {
-            res_facts->type   = obj_facts->decont_type;
-            res_facts->flags |= MVM_SPESH_FACT_KNOWN_TYPE;
+        if (!MVM_spesh_facts_decont_blocked_by_alias(tc, g, ins)) {
+            MVMSpeshFacts *res_facts = MVM_spesh_get_facts(tc, g, ins->operands[0]);
+            int set_facts = 0;
+            if (obj_facts->flags & MVM_SPESH_FACT_KNOWN_DECONT_TYPE) {
+                res_facts->type   = obj_facts->decont_type;
+                res_facts->flags |= MVM_SPESH_FACT_KNOWN_TYPE;
+                set_facts = 1;
+            }
+            if (obj_facts->flags & MVM_SPESH_FACT_DECONT_CONCRETE) {
+                res_facts->flags |= MVM_SPESH_FACT_CONCRETE;
+                set_facts = 1;
+            }
+            else if (obj_facts->flags & MVM_SPESH_FACT_DECONT_TYPEOBJ) {
+                res_facts->flags |= MVM_SPESH_FACT_TYPEOBJ;
+                set_facts = 1;
+            }
+            if (set_facts)
+                MVM_spesh_facts_depend(tc, g, res_facts, obj_facts);
         }
-        if (obj_facts->flags & MVM_SPESH_FACT_DECONT_CONCRETE)
-            res_facts->flags |= MVM_SPESH_FACT_CONCRETE;
-        else if (obj_facts->flags & MVM_SPESH_FACT_DECONT_TYPEOBJ)
-            res_facts->flags |= MVM_SPESH_FACT_TYPEOBJ;
     }
 }
 
@@ -1223,6 +1235,31 @@ static void optimize_extop(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *b
     }
 }
 
+static void optimize_uniprop_ops(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb, MVMSpeshIns *ins) {
+    MVMSpeshFacts *arg1_facts = MVM_spesh_get_facts(tc, g, ins->operands[1]);
+    MVMSpeshFacts *result_facts = MVM_spesh_get_facts(tc, g, ins->operands[0]);
+    if (arg1_facts->flags & MVM_SPESH_FACT_KNOWN_VALUE) {
+        if (ins->info->opcode == MVM_OP_unipropcode) {
+            result_facts->flags |= MVM_SPESH_FACT_KNOWN_VALUE;
+            result_facts->value.i = (MVMint64)MVM_unicode_name_to_property_code(tc, arg1_facts->value.s);
+            ins->info = MVM_op_get_op(MVM_OP_const_i64);
+            ins->operands[1].lit_i64 = result_facts->value.i;
+            arg1_facts->usages--;
+        } else if (ins->info->opcode == MVM_OP_unipvalcode) {
+            MVMSpeshFacts *arg2_facts = MVM_spesh_get_facts(tc, g, ins->operands[2]);
+
+            if (arg2_facts->flags & MVM_SPESH_FACT_KNOWN_VALUE) {
+                result_facts->flags |= MVM_SPESH_FACT_KNOWN_VALUE;
+                result_facts->value.i = (MVMint64)MVM_unicode_name_to_property_value_code(tc, arg1_facts->value.i, arg2_facts->value.s);
+                ins->info = MVM_op_get_op(MVM_OP_const_i64);
+                ins->operands[1].lit_i64 = result_facts->value.i;
+                arg1_facts->usages--;
+                arg2_facts->usages--;
+            }
+        }
+}
+}
+
 /* If something is only kept alive because we log its allocation, kick out
  * the allocation logging and let the op that creates it die.
  */
@@ -1437,6 +1474,7 @@ static void optimize_bb(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb)
         case MVM_OP_prepargs:
             arg_info.cs = g->sf->body.cu->body.callsites[ins->operands[0].callsite_idx];
             arg_info.prepargs_ins = ins;
+            arg_info.prepargs_bb  = bb;
             break;
         case MVM_OP_arg_i:
         case MVM_OP_arg_n:
@@ -1511,6 +1549,10 @@ static void optimize_bb(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb)
         case MVM_OP_objprimspec:
             optimize_objprimspec(tc, g, ins);
             break;
+        case MVM_OP_unipropcode:
+        case MVM_OP_unipvalcode:
+            optimize_uniprop_ops(tc, g, bb, ins);
+            break;
         case MVM_OP_unshift_i:
         case MVM_OP_unshift_n:
         case MVM_OP_unshift_s:
