@@ -70,10 +70,11 @@ static MVMString * re_nfg(MVMThreadContext *tc, MVMString *in) {
     MVMCodepointIter ci;
     MVMint32 ready;
     MVMString *out;
+    MVMuint32 bufsize = in->body.num_graphs;
 
-    /* Create output buffer; it'll never be longer than the initial estimate
-     * since the most we'll do is collapse two things into one in places. */
-    MVMGrapheme32 *out_buffer = MVM_malloc(in->body.num_graphs * sizeof(MVMGrapheme32));
+    /* Create the output buffer. We used to believe it can't ever be bigger
+     * than the initial estimate, but utf8-c8 showed us otherwise. */
+    MVMGrapheme32 *out_buffer = MVM_malloc(bufsize * sizeof(MVMGrapheme32));
     MVMint64 out_pos = 0;
 
     /* Iterate codepoints and normalizer. */
@@ -83,15 +84,27 @@ static MVMString * re_nfg(MVMThreadContext *tc, MVMString *in) {
         MVMGrapheme32 g;
         ready = MVM_unicode_normalizer_process_codepoint_to_grapheme(tc, &norm, MVM_string_ci_get_codepoint(tc, &ci), &g);
         if (ready) {
+            if (out_pos + ready > bufsize) {
+                /* Doubling up the buffer size seems excessive, so just
+                 * add a generous amount of storage */
+                bufsize += ready + 32;
+                out_buffer = MVM_realloc(out_buffer, bufsize * sizeof(MVMGrapheme32));
+            }
             out_buffer[out_pos++] = g;
-            while (--ready > 0)
+            while (--ready > 0) {
                 out_buffer[out_pos++] = MVM_unicode_normalizer_get_grapheme(tc, &norm);
+            }
         }
     }
     MVM_unicode_normalizer_eof(tc, &norm);
     ready = MVM_unicode_normalizer_available(tc, &norm);
-    while (ready--)
+    if (out_pos + ready > bufsize) {
+        bufsize += ready + 1;
+        out_buffer = MVM_realloc(out_buffer, bufsize * sizeof(MVMGrapheme32));
+    }
+    while (ready--) {
         out_buffer[out_pos++] = MVM_unicode_normalizer_get_grapheme(tc, &norm);
+    }
     MVM_unicode_normalizer_cleanup(tc, &norm);
 
     /* Build result string. */
@@ -549,7 +562,7 @@ MVMGrapheme32 MVM_string_ord_at(MVMThreadContext *tc, MVMString *s, MVMint64 off
 
     agraphs = MVM_string_graphs(tc, s);
     if (offset < 0 || offset >= agraphs)
-	return -1;
+        return -1;
 
     g = MVM_string_get_grapheme_at_nocheck(tc, s, offset);
 
@@ -562,24 +575,26 @@ MVMGrapheme32 MVM_string_ord_basechar_at(MVMThreadContext *tc, MVMString *s, MVM
     MVMNormalizer norm;
     MVMint32 ready;
 
-    MVM_string_check_arg(tc, s, "grapheme_at");
+    MVM_string_check_arg(tc, s, "ord_basechar_at");
 
     agraphs = MVM_string_graphs(tc, s);
     if (offset < 0 || offset >= agraphs)
-	return -1;  /* fixes RT #126771 */
+        return -1;  /* fixes RT #126771 */
 
     g = MVM_string_get_grapheme_at_nocheck(tc, s, offset);
 
     if (g < 0) {
-	MVMNFGSynthetic *si = MVM_nfg_get_synthetic_info(tc, g);
-	g = si->base;
+        MVMNFGSynthetic *si = MVM_nfg_get_synthetic_info(tc, g);
+        g = si->base;
     }
     else {
-	MVM_unicode_normalizer_init(tc, &norm, MVM_NORMALIZE_NFD);
-	ready = MVM_unicode_normalizer_process_codepoint_to_grapheme(tc, &norm, g, &g);
-	MVM_unicode_normalizer_eof(tc, &norm);
-	if (!ready)
-	    g = MVM_unicode_normalizer_get_grapheme(tc, &norm);
+        MVM_unicode_normalizer_init(tc, &norm, MVM_NORMALIZE_NFD);
+        ready = MVM_unicode_normalizer_process_codepoint_to_grapheme(tc, &norm, g, &g);
+        MVM_unicode_normalizer_eof(tc, &norm);
+        if (!ready)
+            g = MVM_unicode_normalizer_get_grapheme(tc, &norm);
+
+        MVM_unicode_normalizer_cleanup(tc, &norm);
     }
 
     return g;
@@ -1231,6 +1246,34 @@ void MVM_string_flatten(MVMThreadContext *tc, MVMString *s) {
     }
 }
 
+/* If the string is made up of strands, then produces a flattend string
+ * representing the exact same graphemes but without strands. Otherwise,
+ * returns the input string. Intended for strings that will be indexed
+ * into heavily (when evaluating regexes, for example). */
+MVMString * MVM_string_indexing_optimized(MVMThreadContext *tc, MVMString *s) {
+    MVM_string_check_arg(tc, s, "indexingoptimized");
+    if (s->body.storage_type == MVM_STRING_STRAND) {
+        MVMGrapheme32   *flat = MVM_malloc(MVM_string_graphs(tc, s) * sizeof(MVMGrapheme32));
+        MVMStringStrand *orig = s->body.storage.strands;
+        MVMuint32        i    = 0;
+        MVMString       *res;
+
+        MVMGraphemeIter  gi;
+        MVM_string_gi_init(tc, &gi, s);
+        while (MVM_string_gi_has_more(tc, &gi))
+            flat[i++] = MVM_string_gi_get_grapheme(tc, &gi);
+
+        res = (MVMString *)MVM_repr_alloc_init(tc, tc->instance->VMString);
+        res->body.storage_type    = MVM_STRING_GRAPHEME_32;
+        res->body.storage.blob_32 = flat;
+        res->body.num_graphs      = MVM_string_graphs(tc, s);
+        return res;
+    }
+    else {
+        return s;
+    }
+}
+
 /* Escapes a string, replacing various chars like \n with \\n. Can no doubt be
  * further optimized. */
 MVMString * MVM_string_escape(MVMThreadContext *tc, MVMString *s) {
@@ -1705,17 +1748,17 @@ MVMuint8 MVM_string_find_encoding(MVMThreadContext *tc, MVMString *name) {
     if (!encoding_name_init) {
         MVM_gc_allocate_gen2_default_set(tc);
         encoding_utf8_name        = MVM_string_ascii_decode_nt(tc, tc->instance->VMString, "utf8");
-        MVM_gc_root_add_permanent(tc, (MVMCollectable **)&encoding_utf8_name);
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&encoding_utf8_name, "Encoding name");
         encoding_ascii_name       = MVM_string_ascii_decode_nt(tc, tc->instance->VMString, "ascii");
-        MVM_gc_root_add_permanent(tc, (MVMCollectable **)&encoding_ascii_name);
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&encoding_ascii_name, "Encoding name");
         encoding_latin1_name      = MVM_string_ascii_decode_nt(tc, tc->instance->VMString, "iso-8859-1");
-        MVM_gc_root_add_permanent(tc, (MVMCollectable **)&encoding_latin1_name);
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&encoding_latin1_name, "Encoding name");
         encoding_utf16_name       = MVM_string_ascii_decode_nt(tc, tc->instance->VMString, "utf16");
-        MVM_gc_root_add_permanent(tc, (MVMCollectable **)&encoding_utf16_name);
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&encoding_utf16_name, "Encoding name");
         encoding_windows1252_name = MVM_string_ascii_decode_nt(tc, tc->instance->VMString, "windows-1252");
-        MVM_gc_root_add_permanent(tc, (MVMCollectable **)&encoding_windows1252_name);
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&encoding_windows1252_name, "Encoding name");
         encoding_utf8_c8_name     = MVM_string_ascii_decode_nt(tc, tc->instance->VMString, "utf8-c8");
-        MVM_gc_root_add_permanent(tc, (MVMCollectable **)&encoding_utf8_c8_name);
+        MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&encoding_utf8_c8_name, "Encoding name");
         encoding_name_init   = 1;
         MVM_gc_allocate_gen2_default_clear(tc);
     }
