@@ -1191,7 +1191,7 @@ sub emit_composition_lookup {
     my $u_table_idx = 0;
     my $entries     = '';
     my $l_tables    = 'static const MVMint32 *comp_l_empty[] = {' . ('NULL,' x 256) . "};\n";
-    my $u_tables    = 'static const MVMint32 **comp_u_empty[] = {' . ('NULL,' x 256) . "};\n";
+    my $u_tables    = 'static const MVMint32 **comp_u_empty[] = {' . ('comp_l_empty,' x 256) . "};\n";
     my $p_table     = 'static const MVMint32 ***comp_p[] = {';
     for (my $p = 0; $p < 17; $p++) {
         unless ($lookup[$p]) {
@@ -1627,6 +1627,11 @@ sub tweak_nfg_qc {
             $point->{'NFG_QC'} = 0;
         }
 
+        # Zero Width Joiner
+        elsif ($code == 0x200D) {
+            $point->{'NFG_QC'} = 0;
+        }
+
         # Grapheme_Extend
         elsif ($point->{'Grapheme_Extend'}) {
             $point->{'NFG_QC'} = 0;
